##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""tests of custom widgets defined in Schema"""
import doctest
import re
import unittest

import zope.component.testing
import zope.publisher.interfaces.browser
import zope.schema.interfaces
import zope.traversing.adapters
from zope import component
from zope.formlib.interfaces import ConversionError
from zope.formlib.interfaces import IInputWidget
from zope.formlib.textwidgets import IntWidget
from zope.formlib.textwidgets import TextWidget
from zope.publisher.browser import TestRequest
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.schema import Int
from zope.schema import TextLine
from zope.schema.interfaces import IInt
from zope.schema.interfaces import ITextLine
from zope.schema.interfaces import ValidationError

import zc.form.browser
import zc.form.field
from zc.form.browser.exceptionviews import ValidationErrorView
from zc.form.browser.unionwidget import UnionWidget
from zc.form.field import Union


class TestUnionWidget(
        zope.component.testing.PlacelessSetup, unittest.TestCase):

    def setUp(self):
        super().setUp()
        # XXX cheating: should not rely on these. :-/
        component.provideAdapter(
            TextWidget, (ITextLine, IBrowserRequest), IInputWidget)
        component.provideAdapter(
            IntWidget, (IInt, IBrowserRequest), IInputWidget)
        from zc.form.browser.unionwidget import default_template
        component.provideAdapter(default_template, name='default')
        component.provideAdapter(
            zope.traversing.adapters.DefaultTraversable,
            [None],
        )

    def test_render(self):
        request = TestRequest()
        field = Union(
            (TextLine(title="Name", min_length=5),
             Int(title="Age", min=0)),
            title="Simple Identifier",
            __name__='identifier')
        widget = UnionWidget(field, request)
        widget.setPrefix('field')
        output = widget()
        self.assertIn('<table', output)
        self.assertIn('Age', output)
        self.assertIn('Name', output)
        self.assertRegex(output, r'''type=['"]radio['"]''')
        self.assertRegex(output, r'''name=['"]field.identifier['"]''')
        self.assertRegex(output, r'''id=["']field.identifier-00['"]''')
        self.assertRegex(output, r'''id=["']field.identifier-01['"]''')
        self.assertRegex(
            output, r'''name=["']field.identifier.unioned_00['"]''')
        self.assertRegex(
            output, r'''name=["']field.identifier.unioned_01['"]''')
        self.assertNotRegex(output, r'''id=["']field.identifier-02['"]''')
        self.assertNotRegex(output, r'''checked\s*=\s*['"]checked['"]''')
        field = Union(
            (TextLine(title="Name", min_length=5),
             Int(title="Age", min=0)),
            title="Simple Identifier",
            __name__='identifier',
            required=False)
        widget = UnionWidget(field, request)
        widget.setPrefix('field')
        output = widget()
        self.assertRegex(output, r'''id=["']field.identifier-02['"]''')
        self.assertRegex(output, r'''checked\s*=\s*['"]checked['"]''')

    def test_use_default_for_not_selected(self):
        # test use_default_for_not_selected = True
        request = TestRequest()
        # the default selection shoud be the the option field which has the
        # value of None
        field = Union(
            (zc.form.field.TextLine(
                title="New Password", missing_value='',
                default_getter=lambda x: 'secret password'),
             zc.form.field.Option(
                title="No Change", value='no change')),
            title="Change Password",
            missing_value='',
            use_default_for_not_selected=True,
            __name__='identifier')
        widget = UnionWidget(field, request)
        widget.setPrefix('field')
        output = widget()
        # remove double whitespaces
        normalized_output = " ".join(output.split())

        # the value of the textline field should be the default_getter's
        # result
        value_attr_of_textline = re.search(
            '<input.*id="field.identifier.unioned_00".*(value=".*").*></div>',
            normalized_output).groups()[0]
        self.assertIn('secret password', value_attr_of_textline)

        # the radio button of the option field should be selected
        radio_option_field = re.search(
            '<input.*id="field.identifier-01"(.*)/> </td>',
            normalized_output).groups()[0]
        self.assertIn('checked="checked"', radio_option_field)

    def test_evaluate(self):
        request = TestRequest()
        request.form.update({
            'field.identifier-marker': 'x',
            'field.identifier.unioned_00': 'Foo Bar',
            'field.identifier': '0'})
        field = Union(
            (TextLine(title="Name", min_length=5),
             Int(title="Age", min=0)),
            title="Simple Identifier",
            __name__='identifier')
        widget = UnionWidget(field, request)
        widget.setPrefix('field')
        self.assertEqual(widget.loadValueFromRequest(), 'Foo Bar')


class TestInterfaces(unittest.TestCase):
    """Testing interfaces import."""

    def test_interfaces__import__1(self):
        """zope.form.browser.interfaces can be imported"""
        import zc.form.browser.interfaces  # noqa: F401


class TestValidationErrorView(unittest.TestCase):
    """Testing .exceptionviews.ValidationErrorView."""

    def test_exceptionviews__ValidationErrorView__1(self):
        """It converts an invariant error to an html snippet."""
        err = ValidationError("Bad error!  Bad!")
        view = ValidationErrorView(err, None)
        self.assertEqual(
            view.snippet(), '<span class="error">Bad error!  Bad!</span>')

    def test_exceptionviews__ValidationErrorView__2(self):
        """It converts also unicode an html snippet."""
        err = ValidationError("Fälscher!")
        view = ValidationErrorView(err, None)
        self.assertEqual(
            view.snippet(), '<span class="error">Fälscher!</span>')

    def test_exceptionviews__ValidationErrorView__3(self):
        """It quotes HTML characters correctly."""
        err = ValidationError("The <error> & me.")
        view = ValidationErrorView(err, None)
        self.assertEqual(
            view.snippet(),
            '<span class="error">The &lt;error&gt; &amp; me.</span>')

    def test_exceptionviews__ValidationErrorView__4(self):
        """It converts the exception argument of an ConversionError as well."""
        err = ConversionError(ValidationError("not valid"))
        view = ValidationErrorView(err, None)
        self.assertEqual(
            view.snippet(), '<span class="error">not valid</span>')


def pageSetUp(test):
    zope.component.testing.setUp(test)
    component.provideAdapter(
        zope.traversing.adapters.DefaultTraversable,
        [None],
    )


optionflags = (
    doctest.NORMALIZE_WHITESPACE
    | doctest.ELLIPSIS
    | doctest.REPORT_NDIFF
    | doctest.IGNORE_EXCEPTION_DETAIL
    | doctest.REPORT_ONLY_FIRST_FAILURE
)


def test_suite():
    return unittest.TestSuite([
        unittest.defaultTestLoader.loadTestsFromName(__name__),
        doctest.DocFileSuite(
            '../combinationwidget.rst',
            optionflags=optionflags,
            setUp=pageSetUp,
            tearDown=zope.component.testing.tearDown),
    ])



##############################################################################
#
# Copyright (c) 2003-2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""fields

$Id: field.py 4634 2006-01-06 20:21:15Z fred $
"""
import zope.catalog.interfaces
import zope.index.text.parsetree
import zope.index.text.queryparser
from zope import component
from zope import i18n
from zope import interface
from zope import schema
from zope.interface.exceptions import DoesNotImplement
from zope.schema.interfaces import IField
from zope.schema.interfaces import ValidationError
from zope.schema.interfaces import WrongType

from zc.form import interfaces
from zc.form.i18n import _


_no_unioned_field_validates = _(
    "No unioned field validates ${value}.")

_range_less_error = _("${minimum} must be less than ${maximum}.")
_range_less_equal_error = _(
    "${minimum} must be less than or equal to ${maximum}.")
_combination_wrong_size_error = _("The value has the wrong number of members")
_combination_not_a_sequence_error = _("The value is not a sequence")
_bad_query = _("Invalid query.")

# Union field that accepts other fields...


class MessageValidationError(ValidationError):
    """ValidationError that takes a message"""

    def __init__(self, message, mapping=None):
        if mapping is not None:
            self.message = i18n.Message(message, mapping=mapping)
        else:
            self.message = message
        self.args = (message, mapping)

    def doc(self):
        return self.message


@interface.implementer(interfaces.IExtendedField)
class BaseField(schema.Field):
    """Field with a callable as default and a tuple of constraints.

    >>> def secure_password(field, value):
    ...     if len(value) < 8:
    ...         raise schema.ValidationError('Password too short.')
    ...
    >>> class IDummy(interface.Interface):
    ...     suggested_password = BaseField(
    ...         title=u'Suggested Password',
    ...         default_getter=lambda context: u'asdf',
    ...         constraints=(secure_password, ))
    ...
    >>> f = IDummy['suggested_password'].bind(None) # use None as context
    >>> interfaces.IExtendedField.providedBy(f)
    True
    >>> f.__name__
    'suggested_password'
    >>> print(f.title)
    Suggested Password
    >>> print(f.default)
    asdf
    >>> f.validate(u'123456789')
    >>> f.validate(u'asdf')
    Traceback (most recent call last):
    ...
    ValidationError: Password too short.
    >>> class IDummy2(interface.Interface):
    ...     invalid_default = BaseField(
    ...         title=u'Field with invalid default',
    ...         default=u'standard',
    ...         default_getter=lambda context: u'get default')
    Traceback (most recent call last):
    ...
    TypeError: may not specify both a default and a default_getter
    """
    constraints = ()
    _default = default_getter = None

    def __init__(self, constraints=(), default_getter=None, **kw):
        self.constraints = constraints
        if default_getter is not None and 'default' in kw:
            raise TypeError(
                'may not specify both a default and a default_getter')
        super().__init__(**kw)
        self.default_getter = default_getter

    def _validate(self, value):
        super()._validate(value)
        if value != self.missing_value:
            for constraint in self.constraints:
                constraint(self, value)

    @property
    def default(self):
        if self.default_getter is not None:
            return self.default_getter(self.context)
        else:
            return self._default

    @default.setter
    def default(self, value):
        assert self.default_getter is None
        self._default = value


@interface.implementer(interfaces.IOptionField)
class Option(BaseField):
    """A field with one predefined value."""

    def __init__(self, value=None, value_getter=None,
                 identity_comparison=False, **kw):
        self.value = value
        self.value_getter = value_getter
        self.identity_comparison = identity_comparison
        assert (value is None) ^ (value_getter is None)
        assert not kw.get('required')
        kw['required'] = False
        super().__init__(**kw)

    def _validate(self, value):
        if value != self.missing_value:
            if self.identity_comparison:
                if self.getValue() is not value:
                    raise WrongType
            elif self.getValue() != value:
                raise WrongType

    def getValue(self):
        if self.value_getter is not None:
            return self.value_getter(self.context)
        else:
            return self.value


@interface.implementer(interfaces.IUnionField)
class Union(BaseField):
    """Union field allows a schema field to hold one of many other field types.

    For instance, you might want to make a field that can hold
    a duration *or* a date, if you are working with a PIM app.  Or perhaps
    you want to have a field that can hold a string from a vocabulary *or* a
    custom string.  Both of these examples can be accomplished in a variety of
    ways--the union field is one option.

    The second example is more easily illustrated.  Here is a union field that
    is a simple way of allowing "write-ins" in addition to selections from a
    choice.  We'll be very explicit about imports, in part so this can be
    trivially moved to a doc file test.

    Notice as you look through the example that field order does matter: the
    first field as entered in the field list that validates is chosen as the
    validField; thus, even though the options in the Choice field would also
    validate in a TextLine, the Choice field is identified as the "validField"
    because it is first.

    >>> class IDummy(interface.Interface):
    ...     cartoon_character = Union((
    ...         schema.Choice(
    ...             title=u'Disney',
    ...             description=u'Some tried-and-true Disney characters',
    ...             values=(u'Goofy',u'Mickey',u'Donald',u'Minnie')),
    ...         schema.TextLine(
    ...             title=u'Write-in',
    ...             description=u'Name your own!')),
    ...         required=True,
    ...         title=u'Cartoon Character',
    ...         description=u'Your favorite cartoon character')
    ...
    >>> f = IDummy['cartoon_character']
    >>> interfaces.IUnionField.providedBy(f)
    True
    >>> f.__name__
    'cartoon_character'
    >>> isinstance(f.fields[0], schema.Choice)
    True
    >>> isinstance(f.fields[1], schema.TextLine)
    True
    >>> f.fields[0].__name__ != f.fields[1].__name__
    True
    >>> len(f.fields)
    2
    >>> print(f.title)
    Cartoon Character
    >>> f.validate(u'Goofy')
    >>> f.validField(u'Goofy') is f.fields[0]
    True
    >>> f.validate(u'Calvin')
    >>> f.validField(u'Calvin') is f.fields[1]
    True
    >>> f.validate(42)
    Traceback (most recent call last):
    ...
    MessageValidationError: (u'No unioned field validates ${value}.', {'value': 42})

    That's a working example.  Now lets close with a couple of examples that
    should fall over.

    You must union at least two fields:

    >>> f = Union((schema.TextLine(title=u'Foo Text Line!'),), title=u'Foo')
    Traceback (most recent call last):
    ...
    ValueError: union must combine two or more fields

    And, unsurprisingly, they must actually be fields:

    >>> from zope.interface.exceptions import DoesNotImplement
    >>> try:
    ...     f = Union(('I am not a number.', 'I am a free man!'), title=u'Bar')
    ... except DoesNotImplement:
    ...     print("Not a field")
    ...
    Not a field

    Binding a union field also takes care of binding the contained fields:

    >>> context = object()
    >>> bound_f = f.bind(context)
    >>> bound_f.context is context
    True
    >>> bound_f.fields[0].context is context
    True
    >>> bound_f.fields[1].context is context
    True
    """  # noqa
    fields = ()
    use_default_for_not_selected = False

    def __init__(self, fields, use_default_for_not_selected=False, **kw):
        if len(fields) < 2:
            raise ValueError(_("union must combine two or more fields"))
        for ix, field in enumerate(fields):
            if not IField.providedBy(field):
                raise DoesNotImplement(IField)
            field.__name__ = "unioned_%02d" % ix
        self.fields = tuple(fields)
        self.use_default_for_not_selected = use_default_for_not_selected
        super().__init__(**kw)

    def bind(self, object):
        clone = super().bind(object)
        # We need to bind the fields too, e.g. for Choice fields
        clone.fields = tuple(field.bind(object) for field in clone.fields)
        return clone

    def validField(self, value):
        """Return first valid field, or None."""
        for field in self.fields:
            try:
                field.validate(value)
            except ValidationError:
                pass
            else:
                return field

    def _validate(self, value):
        if self.validField(value) is None:
            raise MessageValidationError(_no_unioned_field_validates,
                                         {'value': value})


class OrderedCombinationConstraint:

    def __init__(self, may_be_equal=True, decreasing=False):
        self.may_be_equal = may_be_equal
        self.decreasing = decreasing

    def __call__(self, field, value):
        # can assume that len(value) == len(field.fields)
        last = None
        for v, f in zip(value, field.fields):
            if v != f.missing_value:
                if last is not None:
                    if self.decreasing:
                        if self.may_be_equal:
                            if v > last:
                                raise MessageValidationError(
                                    _range_less_equal_error,
                                    {'minimum': v, 'maximum': last})
                        elif v >= last:
                            raise MessageValidationError(
                                _range_less_error,
                                {'minimum': v, 'maximum': last})
                    else:
                        if self.may_be_equal:
                            if v < last:
                                raise MessageValidationError(
                                    _range_less_equal_error,
                                    {'minimum': last, 'maximum': v})
                        elif v <= last:
                            raise MessageValidationError(
                                _range_less_error,
                                {'minimum': last, 'maximum': v})
                last = v


@interface.implementer(interfaces.ICombinationField)
class Combination(BaseField):
    """a combination of two or more fields, all of which may be completed.

    It accepts two or more fields.  It also accepts a 'constraints' argument.
    Unlike the usual 'constraint' argument (which is also available), the
    constraints should be a sequence of callables that take a field and a
    value, and they should raise an error if there is a problem.

    >>> from zc.form.field import Combination, OrderedCombinationConstraint
    >>> from zope import schema, interface
    >>> class IDemo(interface.Interface):
    ...     publication_range = Combination(
    ...         (schema.Date(title=u'Begin', required=False),
    ...          schema.Date(title=u'Expire', required=True)),
    ...         title=u'Publication Range',
    ...         required=True,
    ...         constraints=(OrderedCombinationConstraint(),))
    ...
    >>> f = IDemo['publication_range']
    >>> interfaces.ICombinationField.providedBy(f)
    True
    >>> f.__name__
    'publication_range'
    >>> isinstance(f.fields[0], schema.Date)
    True
    >>> isinstance(f.fields[1], schema.Date)
    True
    >>> print(f.title)
    Publication Range
    >>> print(f.fields[0].title)
    Begin
    >>> print(f.fields[1].title)
    Expire
    >>> import datetime
    >>> f.validate((datetime.date(2005, 6, 22), datetime.date(2005, 7, 10)))
    >>> f.validate((None, datetime.date(2005, 7, 10)))
    >>> f.validate((datetime.date(2005, 6, 22), None))
    Traceback (most recent call last):
    ...
    RequiredMissing: combination_01
    >>> f.validate(('foo', datetime.date(2005, 6, 22)))
    Traceback (most recent call last):
    ...
    WrongType: ('foo', <type 'datetime.date'>, 'combination_00')
    >>> f.validate('foo') # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    MessageValidationError: (u'The value has the wrong number of members', ...)
    >>> f.validate(17)
    Traceback (most recent call last):
    ...
    MessageValidationError: (u'The value is not a sequence', None)
    >>> f.validate((datetime.date(2005, 6, 22), datetime.date(1995, 7, 10)))
    ... # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    MessageValidationError: (u'${minimum} must be less than or equal to ...

    Binding a Combination field also takes care of binding contained fields:

    >>> context = object()
    >>> bound_f = f.bind(context)
    >>> bound_f.context is context
    True
    >>> bound_f.fields[0].context is context
    True
    >>> bound_f.fields[1].context is context
    True

    Each entry in the combination has to be a schema field

    >>> class IDemo2(interface.Interface):
    ...     invalid_field = Combination(
    ...         (schema.Date(title=u'Begin', required=False),
    ...          dict(title=u'Expire', required=True)),
    ...         title=u'Invalid field')
    Traceback (most recent call last):
    ...
    DoesNotImplement: An object does not implement interface...
    """
    fields = constraints = ()

    def __init__(self, fields, **kw):
        for ix, field in enumerate(fields):
            if not IField.providedBy(field):
                raise DoesNotImplement(IField)
            field.__name__ = "combination_%02d" % ix
        self.fields = tuple(fields)
        super().__init__(**kw)

    def _validate(self, value):
        if value != self.missing_value:
            try:
                len_value = len(value)
            except (TypeError, AttributeError):
                raise MessageValidationError(
                    _combination_not_a_sequence_error)
            if len_value != len(self.fields):
                raise MessageValidationError(
                    _combination_wrong_size_error)
            for v, f in zip(value, self.fields):
                f = f.bind(self.context)
                f.validate(v)
        super()._validate(value)

    def bind(self, object):
        clone = super().bind(object)
        # We need to bind the fields too, e.g. for Choice fields
        clone.fields = tuple(field.bind(object) for field in clone.fields)
        return clone


class QueryTextLineConstraint(BaseField, schema.TextLine):
    def __init__(self, index_getter=None, catalog_name=None, index_name=None):
        assert not ((catalog_name is None) ^ (index_name is None))
        assert (index_getter is None) ^ (catalog_name is None)
        self.catalog_name = catalog_name
        self.index_name = index_name
        self.index_getter = index_getter

    def __call__(self, field, value):
        if self.index_getter is not None:
            index = self.index_getter(self.context)
        else:
            catalog = component.getUtility(
                zope.catalog.interfaces.ICatalog,
                self.catalog_name,
                field.context)
            index = catalog[self.index_name]
        parser = zope.index.text.queryparser.QueryParser(index.lexicon)
        try:
            parser.parseQuery(value)
        except zope.index.text.parsetree.ParseError:
            raise MessageValidationError(_bad_query)


class TextLine(BaseField, schema.TextLine):
    """An extended TextLine.

    >>> from zope.index.text.textindex import TextIndex
    >>> index = TextIndex()
    >>> class IDemo(interface.Interface):
    ...     query = TextLine(
    ...         constraints=(
    ...             QueryTextLineConstraint(
    ...                 lambda context: index),),
    ...         title=u"Text Query")
    ...
    >>> field = IDemo['query'].bind(None) # using None as context
    >>> field.validate(u'cow')
    >>> field.validate(u'cow and dog')
    >>> field.validate(u'a the') # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    MessageValidationError: (u'Invalid query.', None)
    >>> field.validate(u'and') # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    MessageValidationError: (u'Invalid query.', None)
    >>> field.validate(u'cow not dog')
    >>> field.validate(u'cow not not dog') # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    MessageValidationError: (u'Invalid query.', None)
    >>> field.validate(u'cow and not dog')
    >>> field.validate(u'cow -dog')
    >>> field.validate(b'cow') # non-unicode fails, as usual with TextLine
    Traceback (most recent call last):
    ...
    WrongType: ('cow', <type 'unicode'>, 'query')
    """


@interface.implementer(interfaces.IHTMLSnippet)
class HTMLSnippet(BaseField, schema.Text):
    """Simple implementation for HTML snippet."""


@interface.implementer(interfaces.IHTMLDocument)
class HTMLDocument(BaseField, schema.Text):
    """Simple implementation for HTML document."""



##############################################################################
#
# Copyright (c) 2003-2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Combination widget"""

import zope.cachedescriptors.property
import zope.schema.interfaces
from zope import component
from zope.browserpage import ViewPageTemplateFile
from zope.formlib import namedtemplate
from zope.formlib.interfaces import IDisplayWidget
from zope.formlib.interfaces import IInputWidget
from zope.formlib.interfaces import WidgetInputError

from zc.form.browser.widgetapi import BaseWidget


class CombinationWidget(BaseWidget):

    widget_interface = IInputWidget

    @zope.cachedescriptors.property.Lazy
    def widgets(self):
        field = self.context
        res = []
        for f in field.fields:
            f = f.bind(self.context)
            w = component.getMultiAdapter((f, self.request,),
                                          self.widget_interface)
            w.setPrefix(self.name + ".")
            res.append(w)
        return res

    def setPrefix(self, prefix):
        super().setPrefix(prefix)
        for w in self.widgets:
            w.setPrefix(self.name + ".")

    def loadValueFromRequest(self):
        # the lack of an API to get the input value regardless of validation
        # is a significant problem.  The inability to clear errors is a
        # problem.
        field = self.context
        missing_value = field.missing_value
        widgets = self.widgets
        required_errors = []
        errors = []
        values = []
        any = False
        for w in widgets:
            try:
                val = w.getInputValue()
            except WidgetInputError as e:
                if isinstance(getattr(e, 'errors'),
                              zope.schema.interfaces.RequiredMissing):  # :-(
                    required_errors.append((w, e))
                else:
                    errors.append((w, e))
                val = w.context.missing_value
            values.append(val)
            any = any or val != w.context.missing_value
        if field.required or any or errors:
            errors.extend(required_errors)
        else:  # remove the required errors in the sub widgets
            for w, e in required_errors:
                w.error = lambda: None  # :-(
        if errors:
            if len(errors) == 1:
                errors = errors[0][1]
            else:
                errors = [e for widget, e in errors]
            self._error = WidgetInputError(
                self.context.__name__,
                self.label, errors)
            values = missing_value
        elif not any:
            values = missing_value
        else:
            values = tuple(values)
        return values

    template = namedtemplate.NamedTemplate('default')

    def render(self, value):
        field = self.context
        missing_value = field.missing_value
        if value is not missing_value:
            try:
                len_value = len(value)
            except (TypeError, AttributeError):
                value = missing_value
                self._set_values_on_widgets(value)
            else:
                if len_value != len(field.fields):
                    value = missing_value
                    self._set_values_on_widgets(value)
        if value is not missing_value:
            self._set_values_on_widgets(value)
        for w in self.widgets:  # XXX quick hack.
            if zope.schema.interfaces.IBool.providedBy(w.context):
                w.invert_label = True
            else:
                w.invert_label = False
        return self.template()

    def _set_values_on_widgets(self, values):
        hasInput = self.hasInput()
        for w, v in zip(self.widgets, values):
            if not hasInput or v != w.context.missing_value:
                w.setRenderedValue(v)


default_template = namedtemplate.NamedTemplateImplementation(
    ViewPageTemplateFile('combinationwidget.pt'), CombinationWidget)


class CombinationDisplayWidget(CombinationWidget):

    widget_interface = IDisplayWidget



import pytz

import zc.sourcefactory.basic


class USTimeZones(zc.sourcefactory.basic.BasicSourceFactory):
    """List of timezones taken from pytz"""

    tzs = [tz
           for tz in pytz.common_timezones
           if 'US/' in tz and 'Pacific-New' not in tz]

    def getValues(self):
        return self.tzs

    def getTitle(self, value):
        return value

    def getToken(self, value):
        return value



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Views for exceptions used by the Schema product."""

from html import escape

from zope import component
from zope import i18n
from zope import interface
from zope.exceptions.interfaces import UserError
from zope.formlib.interfaces import ConversionError
from zope.formlib.interfaces import IWidgetInputErrorView
from zope.publisher.interfaces.browser import IBrowserRequest
from zope.schema.interfaces import ValidationError


@interface.implementer(IWidgetInputErrorView)
class AbstractErrorView:
    """Base error view."""

    def __init__(self, context, request):
        self.context, self.request = context, request

    def snippet(self):
        """Convert an invariant error to an html snippet."""
        msg = self.context.args[0]
        msg = i18n.translate(msg, context=self.request, default=msg)
        return '<span class="error">%s</span>' % escape(str(msg))


@component.adapter(ValidationError, IBrowserRequest)
class ValidationErrorView(AbstractErrorView):
    """An error view for ValidationError."""


@component.adapter(ConversionError, IBrowserRequest)
class ConversionErrorView(AbstractErrorView):
    """An error view for ConversionError."""


@component.adapter(UserError, IBrowserRequest)
class UserErrorView(AbstractErrorView):
    """An error view for UserErrorError."""



##############################################################################
#
# Copyright (c) 2003-2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""interfaces.py: interfaces for the form browser machinery

$Id: interfaces.py 920 2005-03-10 00:20:26Z benji $"""

from zope import interface


class ISchemaView(interface.Interface):
    "the basic schema view interface suitable for extension and for display"

    extra = interface.Attribute(
        """A dictionary to store values desired by a given schema view.
        The add view is the only standard interface that explicitly declares a
        use for this dictionary.""")

    buttons = interface.Attribute(
        """A dictionary of a list of dictionaries,
        representing the available buttons.

        The top level dictionary has keys of unicode names representing
        button categories.  The default category defined in the buttons
        subdirective, for example, is u'main'.  The values are lists of
        dictionaries. Each dictionary represents a button within the category
        and contains the keys 'label,', which should be used to draw the
        button; 'id,' which should be used as the name of the submit button
        and typically as the id of the widget on the form (for DOM
        getElementById); 'action,' the name of the method on this class to
        which __call__ should dispatch if the id is activated; 'css_class,'
        the name of the css_class that the button should be assigned;
        'category', the name of the containing category, as described above;
        and 'active,' a boolean indicating if the action is currently
        available for this context (user, condition, etc.).  If a button is
        not active, the id and action are empty strings.  Button dictionaries
        may contain additional keys and values as form subclasses desire.

        Typically, the button_views method is used to render buttons, but
        the buttons property provides a more raw presentation of the
        information.

        All values are essentially advisory except for 'action' and 'id'.  'id'
        is typically auto-generated in such a manner as to minimize the
        likelihood of conflicting with other names on the form.
        """)

    label = interface.Attribute(
        """The suggested label for the form, as might be rendered in a header
        at the top of the page.""")

    fieldNames = interface.Attribute(
        "list of field names in suggested order for the form")

    active_button = interface.Attribute(
        """after the form machinery has determined the button pressed, if any,
        and before the button action has been dispatched, this holds the button
        dictionary, as decribed in the buttons attribute above, of the button
        that was submitted.""")

    update_status = interface.Attribute(
        """The status message, as translated from the status returned from
        the dispatched action.""")

    schema = interface.Attribute(
        """The schema that is the source of this form.""")

    errors = interface.Attribute(
        """a tuple of the error objects generated by the dispatched action
        for each field.""")

    invariant_errors = interface.Attribute(
        """a tuple of the error objects generated by the dispatched action
        for each invariant rule.""")

    widgets = interface.Attribute(
        """After setUpWidgets is called, this contains a collection of widgets
        with a mapping interface.  Iterating over it is in the order of the
        view's fieldNames.  Widgets can be obtained by field name using
        __getitem__ or get.
        """)

    def setUpWidgets():
        """the call to set up widgets for the view.  After this call, the
        widgets attribute is in place.  Usually called in view
        initialization"""

    def setPrefix(prefix):
        "Set the desired form prefix on all widgets."

    def invariant_error_views(self):
        """return a view of each error object in the invariant_errors
        attribute, as found for the 'snippet' name."""

    def error_views():
        """return a view of each error object in the errors
        attribute, as found for the 'snippet' name."""

    def button_views(category=None, show_disabled=False):
        """return an iterator of standard html submit buttons for each button
        definition. If category is None (default), all buttons in all
        categories are rendered; otherwise the value is used to look up and
        only render buttons in the specified category.  A KeyError is raised
        if the category does not exist.  By default, disabled buttons
        (buttons that are not active) are not rendered; if show_disabled is
        True, these are rendered as disabled buttons.
        """

    def beforeRender(self):
        """hook for performing things before rendering.  Note this is called
        whether or not the form is actually rendered: check self.halt_rendering
        if you need to perform different behavior in this circumstance."""

    def __call__(template_usage='', *args, **kw):
        """a dispatcher to button actions and renderer of the form.

        __call__ searches the request for each of the ids in the buttons
        attribute.  As soon as it finds one, it dispatches to the associated
        action, assuming there will be only one action per submission.  The
        active button is stored as active_button, and then the beforeRender
        hook is called.  if halt_rendering has been set to a Boolean True
        value, rendering is stopped and an empty unicode string is returned.
        Otherwise, the return value of the dispatch is (translated to be) the
        update_status, and the form is rendered with the passed arguments.
        """

    def validateInvariants(obj):
        """Given an object, validates the schema invariants and returns a
        list of any generated errors."""

    def getPseudoObject(data):
        """returns an object that claims to implement the form's schema and
        has the data supplied stored on the object.  A helper for validating
        invariants."""

    halt_rendering = interface.Attribute(
        """if halt_rendering is set to True by a dispatched action in
        __call__,  the form should not render.""")

    label_msgid = interface.Attribute(
        """the message id of the label that should  typically be rendered at
        the top of this form.""")

    label_title = interface.Attribute(
        """substituted into label_msgid if the message id translation
        includes a "title" substitution slot (a la "Add ${title}").""")

    raw_buttons = interface.Attribute(
        """the default source of the buttons attribute. It is a  dictionary of
        button group id to sequence of dicts, in which each dict has keys id,
        label, action, permission, condition, and css_class.  action is the
        name of a method on the (composite, with optional bases provided in
        zcml) view class; the value returned is set as the update_status.  The
        user must have the provided permission, if any, and the condition must
        pass, in order for the action to be available as a button.  The
        condition is a TALES expression with 'context', 'adapted', 'nothing',
        'request', and 'modules' available.""")

    # form.pt in this product uses this hook.  This functionality is
    # exercised by forms in other products.
    extra_script = interface.Attribute(
        """a hook point for schema views to suggest arbitrary
        javascript to be rendered in the form.""")


class IEditView(ISchemaView):
    "the basic edit view interface"
    def attempt_commit():
        """The standard action method for an edit view's submit button: will
        validate the entered data and, if it passes, make the changes.
        """

    def commit_changes():
        """If no errors are found in attempt_commit, commit_changes is called
        to actually edit the object.  Override to change what "edit the object"
        means.
        """

    def changed():
        """A hook.  Override to execute logic *after* changes have been made
        successfully."""


class IAddView(ISchemaView):
    "the basic add view interface"

    extra = interface.Attribute(
        """The standard extra-value dictionary described in ISchemaView must
        have "content_factory" as a key to the factory (such as a class) that
        instantiates the object.  It must also have one or more of the keys
        "arguments", "keyword_arguments", "set_before_add", "set_after_add"
        with names of arguments to be used for creation.
        """)

    def attempt_commit():
        """The standard action method for an add view's submit button: will
        validate the entered data and, if it passes, create and add the object.
        """

    def perform_cancel():
        """An action to cancel the add and redirect to the add's container or
        to the previous add form for nested adds.
        """

    def abstract_attempt_commit():
        """The guts of attempting a commit; does everything except redirect, so
        you can override attempt_commit and redirect as desired."""

    def createAndAdd(data):
        """Add the desired object using the data in the data argument.

        The data argument is a dictionary with the data entered in the form.

        returns created object.
        """

    def create(*args, **kw):
        "actually instantiates object and returns it"

    def add(content):
        "adds the content; typically delegates to the adding."

    def changed():
        """A hook.  Override to execute logic *after* an object has been
        created and added successfully."""

    def nextURL():
        "returns the next URL; usually defers to the adding."



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""timezone widget"""

import pytz
# XXX argh.
import pytz.tzinfo

import zope.browserpage
import zope.formlib.interfaces
import zope.interface.common.idatetime
import zope.publisher.interfaces.browser
from zope import component
from zope import interface
from zope.interface.common.idatetime import ITZInfo
from zope.security.checker import NamesChecker

import zc.form.browser.mruwidget
import zc.form.interfaces


ALL_TIMEZONES = frozenset(pytz.all_timezones)

names = set(ITZInfo.names(all=True))
names.add('zone')
names.add('localize')
checker = NamesChecker(names)
pytz.UTC.__Security_checker__ = checker
pytz.tzinfo.BaseTzInfo.__Security_checker__ = checker
# end argh.


class TimeZoneWidget(zc.form.browser.mruwidget.MruSourceInputWidget):

    def getMostRecentlyUsedTerms(self):
        mru = super().getMostRecentlyUsedTerms()
        # add ones from locale
        territory = self.request.locale.id.territory
        if territory:
            try:
                choices = pytz.country_timezones(territory)
            except KeyError:
                pass
            else:
                already = {term.token for term in mru}
                additional = sorted(t for t in choices if t not in already)
                mru.extend(zc.form.interfaces.Term(t.replace('_', ' '), t)
                           for t in additional)
        return mru


@component.adapter(zc.form.interfaces.AvailableTimeZones,
                   zope.publisher.interfaces.browser.IBrowserRequest)
@interface.implementer(zope.formlib.interfaces.ISourceQueryView)
class TimeZoneQueryView:

    def __init__(self, source, request):
        self.context = source
        self.request = request

    _render = zope.browserpage.ViewPageTemplateFile('timezone_queryview.pt')

    def render(self, name):
        return self._render(field_name=name + '.searchstring',
                            button_name=name + '.search')

    def results(self, name):
        if not (name + '.search' in self.request):
            return None

        searchstring = self.request[name + '.searchstring'].lower()
        timezones = []
        searchstring = searchstring.strip().lower().replace(' ', '_')  # regex
        for tz in ALL_TIMEZONES:
            if searchstring in tz.lower():
                timezones.append(pytz.timezone(tz))
        return timezones



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""functional test harness for zc.mruwidget"""
import doctest
import unittest

import zope.app.wsgi.testlayer
import zope.browser.interfaces
import zope.formlib.interfaces
import zope.schema.interfaces
from zope import component
from zope import interface

import zc.form.browser


colors = ['red', 'green', 'cerulean blue']


@interface.implementer(zope.schema.interfaces.ISource)
class AvailableColors:

    def __contains__(self, value):
        return value in colors


@interface.implementer(zope.schema.interfaces.ITitledTokenizedTerm)
class Term:

    def __init__(self, title, token):
        self.title = title
        self.token = token


@component.adapter(zope.schema.interfaces.ISource,
                   zope.publisher.interfaces.browser.IBrowserRequest)
@interface.implementer(zope.browser.interfaces.ITerms)
class ColorTerms:
    """Term and value support needed by query widgets."""

    def __init__(self, source, request):
        self.request = request
        self.source = source

    def getTerm(self, value):
        if value not in self.source:
            raise LookupError
        token = value + '_token'
        title = value.capitalize()
        return Term(title, token)

    def getValue(self, token):
        return token.split('_')[0]


@component.adapter(zope.schema.interfaces.ISource,
                   zope.publisher.interfaces.browser.IBrowserRequest)
@interface.implementer(zope.formlib.interfaces.ISourceQueryView)
class SimpleColorSourceQueryView:

    def __init__(self, source, request):
        self.context = source
        self.request = request

    def render(self, name):
        return '<query view for colors>'

    def results(self, name):
        if not (name + '.search' in self.request):
            return None
        searchstring = self.request[name + '.searchstring']
        return [c for c in colors if searchstring in c]


ZCFormLayer = zope.app.wsgi.testlayer.BrowserLayer(zc.form.browser)


def setUp(test):
    component.provideAdapter(ColorTerms)
    component.provideAdapter(SimpleColorSourceQueryView)


def test_suite():
    # This function must return something test case/suite like to keep
    # zope.testrunner happy.
    tests = unittest.TestSuite()
    try:
        import zc.resourcelibrary  # noqa
    except ModuleNotFoundError:
        # We want to skip the mruwidget tests
        pass
    else:
        suite = doctest.DocFileSuite(
            "../mruwidget.rst",
            globs={'AvailableColors': AvailableColors(),
                   'getRootFolder': ZCFormLayer.getRootFolder},
            optionflags=(doctest.NORMALIZE_WHITESPACE
                         | doctest.ELLIPSIS
                         | doctest.REPORT_ONLY_FIRST_FAILURE
                         | doctest.IGNORE_EXCEPTION_DETAIL
                         | doctest.REPORT_NDIFF),
            setUp=setUp,
        )
        suite.layer = ZCFormLayer
        tests = suite
    return tests



import os.path

from setuptools import setup


def read(filepath):
    with open(os.path.join(*filepath.split('/'))) as f:
        return f.read()


MRU_REQUIRES = [
    'zope.annotation',
    'BTrees',
    'persistent',
    'zc.resourcelibrary',
]

TEST_REQUIRES = [
    'zope.app.appsetup',
    'zope.app.principalannotation',
    'zope.app.wsgi[testlayer] >= 5.2',
    'zope.configuration',
    'zope.container',
    'zope.testing',
    'zope.testrunner >= 6.4',
    'zope.traversing',
]


setup(
    name="zc.form",
    version='3.0',
    url='https://github.com/zopefoundation/zc.form',
    author='Zope Foundation and Contributors',
    author_email='zope-dev@zope.dev',
    include_package_data=True,
    description=(
        'Extra browser widgets and alternative approaches for zope.formlib.'),
    long_description='\n\n'.join([
        read('README.rst'),
        '.. contents::',
        read('CHANGES.rst'),
        read('src/zc/form/browser/combinationwidget.rst'),
        read('src/zc/form/browser/mruwidget.rst'),
    ]),
    license="ZPL-2.1",
    keywords="zope formlib form widget extra",
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Zope Public License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Topic :: Internet :: WWW/HTTP',
        'Framework :: Zope :: 3',
    ],
    python_requires='>=3.9',
    install_requires=[
        'pytz',
        'setuptools',
        'zc.sourcefactory',
        'zope.browserpage',
        'zope.cachedescriptors',
        'zope.catalog',
        'zope.component >= 3.8',
        'zope.exceptions',
        'zope.formlib >= 4.0',
        'zope.index',
        'zope.interface',
        'zope.publisher',
        'zope.schema >= 3.6',
        'zope.security',
    ],
    extras_require=dict(
        mruwidget=MRU_REQUIRES,
        test=MRU_REQUIRES + TEST_REQUIRES,
        slimtest=TEST_REQUIRES,
    ),
    zip_safe=False
)
